import inspect
import subprocess as subproc
import sys
import os
import string
from json import loads, dumps
from typing import (
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    Any,
    Iterable,
    Generator,
    Callable,
    Optional,
)

import random

ALPHANUMERIC_CHARS = string.ascii_letters + string.digits
USERNAME_ASCII_CHARS = ALPHANUMERIC_CHARS + '_-'


def nlprint(*args, **kwargs) -> None:
    """Identical to normal print() but also prints an extra \n above/below"""
    print()
    print(*args, **kwargs)
    print()


def is_jwt_str(s: str) -> bool:
    """Check if a string is an encrypted JWT"""
    parts = str(s).split('.')
    return len(parts) == 3


def ranstr(
    min_length: int = 16,
    max_length: int = 32,
    chars: Iterable = ALPHANUMERIC_CHARS,
    as_generator: bool = False,
) -> Union[Generator, str]:
    """Generate a random string of specified length and characters between min_length and max_length.

    Args:
        min_length (int, optional): The length of the string. Defaults to 16.
        max_length (int, optional): The maximum length of the string. Defaults to 32.
        chars (Iterable, optional): The characters to be used for generating the string.
            Defaults to ALPHANUMERIC_CHARS.
        as_generator (bool, optional): Specifies whether to return a generator instead of a string.
            Defaults to False.
    Returns:
        Union[Generator, str]: The generated string if `as_generator` is False.
            A generator if `as_generator` is True.
    """
    length = random.randint(min_length, max_length)
    if as_generator:
        return (random.choice(chars) for _ in range(length))
    return ''.join(random.choice(chars) for _ in range(length))


def runcmd(
    cmd: str, output: bool = True, *args, **kwargs
) -> Optional[List[str]]:
    """
    Run a command in the shell.

    Args:
        cmd (str): The command to be executed.
        output (bool, optional): Specifies whether to capture and return the output of the command.
            Defaults to True.
        *args: Additional positional arguments to be passed to subprocess.run().
        **kwargs: Additional keyword arguments to be passed to subprocess.run().

    Returns:
        Optional[List[str]]: The captured output of the command as a list of lines if `output` is True.
            None if `output` is False.

    Raises:
        CalledProcessError: If the command exits with a non-zero status and `check=True`.

    """
    if output:
        return subproc.run(
            [c for c in cmd.split()],
            check=True,
            text=True,
            capture_output=True,
            *args,
            **kwargs,
        ).stdout.splitlines()
    else:
        subproc.run(
            [c for c in cmd.split()],
            check=False,
            text=False,
            capture_output=False,
            *args,
            **kwargs,
        )


def get_terminal_width(default: int = 80) -> int:
    """Returns the terminal width. If not detectable, returns default value."""
    try:
        return os.get_terminal_size().columns
    except OSError:
        return default


def print_middle(obj: any, char: str = '=', noprint: bool = False) -> str:
    """
    Print the string representation of an object centered in a line of characters
    based on the terminal width.

    Args:
        obj (Any): The object whose string representation needs to be printed.
        char (str, optional): The character used for filling. Defaults to '='.
        noprint (bool, optional): Specifies whether to avoid printing the generated string.
            Defaults to False.

    Returns:
        str: The generated string.
    """
    terminal_width = get_terminal_width()
    obj_str = str(obj)

    # Calculate the padding on each side. If the total length is odd,
    # the extra space will be on the right side.
    padding = (terminal_width - len(obj_str) - 2) // 2

    result = char * padding + ' ' + obj_str + ' ' + char * padding

    if not noprint:
        print(result)
    return result


def print_columns(iterable: Iterable, separator: str = "  ") -> None:
    objs = list(sorted(iterable))  # Convert any iterable to a list
    terminal_width = get_terminal_width()

    # Calculate max length of str representation of all objects
    max_length = max(len(str(obj)) for obj in objs)

    # Calculate the width of one column, including its separator
    column_width = max_length + len(separator)

    # Calculate number of columns that can fit in terminal width
    num_columns = terminal_width // column_width

    # Adjust if terminal_width is smaller than max_length
    num_columns = max(1, num_columns)

    # Calculate the number of rows needed
    num_rows = -(-len(objs) // num_columns)  # ceiling division

    # Print the objects in rows
    for row in range(num_rows):
        line = []
        for col in range(num_columns):
            index = (
                row * num_columns + col
            )  # This line is correct for row-first printing
            if index < len(objs):
                line.append(str(objs[index]).ljust(max_length))
        print(separator.join(line))


def objinfo(obj: Any):
    """Print information about an object."""
    terminal_width = get_terminal_width()

    obj_type = type(obj)
    obj_name = obj.__name__ if hasattr(obj, "__name__") else "N/A"
    obj_attrs = sorted(
        str(attr) for attr in dir(obj) if not callable(getattr(obj, attr))
    )
    obj_methods = sorted(
        str(method) for method in dir(obj) if callable(getattr(obj, method))
    )
    obj_doc = inspect.getdoc(obj) or "No documentation available."
    obj_scope = (
        "Local"
        if obj in inspect.currentframe().f_back.f_locals.values()
        else "Global"
    )
    obj_size = sys.getsizeof(obj)
    obj_mutability = "Mutable" if hasattr(obj, "__dict__") else "Immutable"
    obj_identity = id(obj)

    def print_info():
        header = f"{obj_name} ({obj_type})"
        subheader = (
            f"Size: {obj_size}",
            f"Scope: {obj_scope}",
            f"Mutability: {obj_mutability}",
            f"Identity: {obj_identity}",
        )
        print()
        print_middle(header)
        print()
        print_columns(subheader)
        nlprint(f'{print_middle("Attributes",  "-", noprint=True)}')
        print_columns(obj_attrs)
        nlprint(f'{print_middle("Methods", "-", noprint=True)}')
        print_columns(obj_methods)
        nlprint(f'{print_middle("Documentation", "-", noprint=True)}')
        print(obj_doc)
        nlprint('-' * terminal_width)

    print_info()


def default_repr(
    obj: Any, transform: Optional[Callable] = None, *args, **kwargs
) -> str:
    """
    Return a string representation of a custom Python object.

    This representation is constructed such that the object can be
    reconstructed from the returned string.

    Args:
        obj (Any): The input Python object.
        transform (Optional[Callable]): A function that will return the repr
            string with modifications. *args/**kwargs will be passed to this function.

    Returns:
        str: The string representation of the object.
    """
    # If the object has a __dict__ attribute, use that
    if hasattr(obj, '__dict__'):
        attributes = ', '.join(
            f"{key}={repr(value)}"
            for key, value in obj.__dict__.items()
            if not hasattr(value, '__call__')
            and not str(key).startswith("_")
        )
    # Otherwise, use dir() to gather potential attributes
    else:
        if isinstance(obj, int):
            return 'int(%s)' % obj
        elif isinstance(obj, float):
            return 'float(%s)' % obj
        elif isinstance(obj, str):
            return 'str(%s)' % obj
        elif isinstance(obj, set):
            return 'set(%s)' % obj
        elif isinstance(obj, (list, tuple, dict)):
            return str(obj)

        attributes = ', '.join(
            f"{attr}={getattr(obj, attr)}"
            for attr in dir(obj)
            if not hasattr(obj, '__call__') and not str(attr).startswith('_')
        )

    if transform is not None:
        return transform(obj, *args, **kwargs)
    return f"{obj.__class__.__name__}({attributes})"



import unittest
from subprocess import CompletedProcess
import sys
import os
from unittest.mock import patch, MagicMock, call
from os.path import dirname, abspath
from typing import Generator


MFROOT = str(dirname(dirname(abspath(__file__))))
if MFROOT not in sys.path:
    sys.path.insert(0, MFROOT)

from myfuncs import (
    runcmd,
    objinfo,
    nlprint,
    is_jwt_str,
    ranstr,
    get_terminal_width,
    print_middle,
    print_columns,
    default_repr,
    ALPHANUMERIC_CHARS,
)


class TestRunCmd(unittest.TestCase):
    def test_runcmd_with_output(self):
        # Mock the subprocess.run() function to return a CompletedProcess object
        mock_completed_process = CompletedProcess(
            args=['echo', 'Hello, World!'],
            returncode=0,
            stdout='Hello, World!\n',
            stderr='',
        )
        with patch('subprocess.run', return_value=mock_completed_process):
            result = runcmd('echo Hello, World!')

        self.assertEqual(result, ['Hello, World!'])

    def test_runcmd_without_output(self):
        # Mock the subprocess.run() function to return None
        with patch('subprocess.run'):
            result = runcmd('echo Hello, World!', output=False)

        self.assertIsNone(result)


class TestPrintMiddle(unittest.TestCase):
    @patch("myfuncs.get_terminal_width", return_value=50)
    def test_print_middle_no_print(self, *args):
        tstr = '012345678901'
        clen = (50 - len(f' {tstr} ')) // 2
        with patch('builtins.print') as mock_print:
            result = print_middle(tstr, noprint=True)
        self.assertEqual(result, f'{clen * "="} {tstr} {clen * "="}')
        mock_print.assert_not_called()


class TestGetTerminalWidth(unittest.TestCase):
    def test_default_terminal_width(self):
        # When there's an OSError exception
        with patch('os.get_terminal_size', side_effect=OSError):
            result = get_terminal_width()
        self.assertEqual(result, 80)

    def test_actual_terminal_width(self):
        with patch(
            'os.get_terminal_size', return_value=os.terminal_size((100, 24))
        ):
            result = get_terminal_width()
        self.assertEqual(result, 100)


class TestPrintColumns(unittest.TestCase):
    items = [
        'a',
        'aa',
        'aaa',
        'aaaa',
        'aaaaa',
        'aaaaaa',
        'aaaaaaa',
        'aaaaaaaa',
        'aaaaaaaaa',
    ]

    @patch("myfuncs.get_terminal_width", return_value=80)
    @patch("builtins.print")
    def test_print_columns_basic(self, mock_print, mock_get_terminal_width):
        print_columns(self.items)
        mock_print.assert_called_with(
            'aaaaaaaa   aaaaaaaaa'
        )  # This verifies the last call to print function

    @patch("myfuncs.get_terminal_width", return_value=40)
    @patch("builtins.print")
    def test_print_columns_small_terminal(
        self, mock_print, mock_get_terminal_width
    ):
        print_columns(self.items)

        mock_print.assert_called_with(
            'aaaaaaa    aaaaaaaa   aaaaaaaaa'
        )  # This verifies the last call to print function

    @patch("myfuncs.get_terminal_width", return_value=80)
    @patch("builtins.print")
    def test_print_columns_sorted(self, mock_print, mock_get_terminal_width):
        print_columns(['b', 'a', 'c'])

        mock_print.assert_called_with(
            'a  b  c'
        )  # This verifies the last call to print function


class TestRanStr(unittest.TestCase):
    def test_string_length(self):
        result = ranstr(5, 10)
        self.assertTrue(5 <= len(result) <= 10)

    def test_string_characters(self):
        chars = 'abcd'
        result = ranstr(5, 10, chars=chars)
        for char in result:
            self.assertIn(char, chars)

    def test_return_generator(self):
        result = ranstr(5, 10, as_generator=True)
        self.assertIsInstance(result, Generator)
        generated_string = ''.join(result)
        self.assertTrue(5 <= len(generated_string) <= 10)

    def test_return_string(self):
        result = ranstr(5, 10)
        self.assertIsInstance(result, str)
        self.assertTrue(5 <= len(result) <= 10)


class TestCustomReprFunction(unittest.TestCase):
    class MyClass:
        def __init__(self, a, b):
            self.a = a
            self.b = b

        def method(self):
            pass

    class AnotherClass:
        c = "class attribute"

        def __init__(self, x):
            self.x = x

    def test_simple_class(self):
        instance = self.MyClass(1, "test")
        representation = default_repr(instance)
        expected_repr = "MyClass(a=1, b='test')"
        self.assertEqual(representation, expected_repr)

    def test_class_with_class_attribute(self):
        instance = self.AnotherClass(5)
        representation = default_repr(instance)
        expected_repr = "AnotherClass(x=5)"
        self.assertEqual(representation, expected_repr)

    def test_builtin_type_without_dict(self):
        value = 123
        representation = default_repr(value)
        expected_repr = "int(123)"
        self.assertEqual(representation, expected_repr)

    def test_list(self):
        lst = [1, 2, 3]
        representation = default_repr(lst)
        expected_repr = "[1, 2, 3]"
        self.assertEqual(representation, expected_repr)

    #def test_nested_class(self):
    #    class Nested:
    #        def __init__(self, y):
    #            self.y = y
#
    #    instance = self.MyClass(1, Nested("inner"))
    #    representation = default_repr(instance)
    #    expected_repr = "MyClass(a=1, b=Nested(y='inner'))"
    #    self.assertEqual(representation, expected_repr)

    def test_set(self):
        st = {1, 2, 3}
        representation = default_repr(st)
        expected_repr = "set({1, 2, 3})"
        self.assertEqual(representation, expected_repr)

    def test_float(self):
        value = 123.45
        representation = default_repr(value)
        expected_repr = "float(123.45)"
        self.assertEqual(representation, expected_repr)

    def test_tuple(self):
        tpl = (1, 2, 3)
        representation = default_repr(tpl)
        expected_repr = "(1, 2, 3)"
        self.assertEqual(representation, expected_repr)

    def test_dict(self):
        d = {'a': 1, 'b': 2}
        representation = default_repr(d)
        expected_repr = "{'a': 1, 'b': 2}"
        self.assertEqual(representation, expected_repr)

    def test_with_private_attribute(self):
        class WithPrivate:
            def __init__(self, x):
                self._x = x

        instance = WithPrivate(5)
        representation = default_repr(instance)
        expected_repr = "WithPrivate()"
        self.assertEqual(representation, expected_repr)


class TestDefaultReprFunction(unittest.TestCase):
    class MockCAPTCHA:
        def __init__(self):
            self.__dict__ = {
                '__module__': 'flask_simple_captcha.captcha',
                '__init__': lambda: None,
                '__repr__': lambda: None,
                '__dict__': {},
                '__weakref__': None,
                '__doc__': None
            }

    def test_mock_captcha_repr(self):
        instance = self.MockCAPTCHA()
        representation = default_repr(instance)
        self.assertIsInstance(representation, str)


if __name__ == '__main__':
    unittest.main()
